﻿using System;

namespace SimplexMethod
{
    /// <summary>
    /// Simplex - trida implementujici Simplexovou Metodu
    /// </summary>
    public class Simplex
    {
        private double[,] tableau; // Simplex "Tableau" - matice rovnic + jednotkova matice
        private int numberOfConstraints; // omezujici (ne)rovnice
        private int numberOfOriginalVariables; // ucelove fce Z

        private SimplexType type;

        private int[] basis; // bazicka promenna

        public Simplex(double[,] tableau, int numberOfConstraint, int numberOfOriginalVariable, SimplexType type)
        {
            this.type = type; // MIN/MAX
            this.numberOfConstraints = numberOfConstraint; // omezujici (ne)rovnice
            this.numberOfOriginalVariables = numberOfOriginalVariable; // ucelova fce Z
            this.tableau = tableau; // matice

            basis = new int[numberOfConstraints];
            for (int i = 0; i < numberOfConstraints; i++)
            {
                basis[i] = numberOfOriginalVariables + i;
            }
        }

        /// <summary>
		/// Vyres Simplexovu Metodu
		/// </summary>
        public void solve()
        {
            while (true)
            {
                printStep();
                int column = 0;

                // najdi klicovy sloupec - "Pivot Column"
                if (this.type.Equals(SimplexType.MAXIMAZE)) // pokud MAXimalizuji
                {
                    column = maxi(); // najdi klicovy sloupec - "pivot column" s nejvice pozitivni cenou
                }
                else // SimplexType.MINIMAZE // pokud MINimalizuji
                {
                    column = mini();  // najdi klicovy sloupec - "pivot column" s nejmene pozitivni cenou
                }
                if (column < 0)
                    break; // optimalni

                // najdi klicovy radek - "pivot Row"
                int row = minRatioRule(column);
                if (row < 0)
                    throw new ArithmeticException("Linearni program neni omezen");

                // klicovy prvek - "pivot"
                pivot(row, column);

                // aktualizuj bazickou promennou
                basis[row] = column;
            }
        }

        /// <summary>
		/// Metoda hledajici index klicoveho sloupce s nejvice pozitivni cenou - pivot column
		/// </summary>
		/// <returns>index klicoveho sloupce s nejvice pozitivni cenou nebo -1</returns>
        private int maxi()
        {
            int column = 0;
            for (int i = 1; i < numberOfConstraints + numberOfOriginalVariables; i++)
            {
                if (tableau[numberOfConstraints, i] > tableau[numberOfConstraints, column])
                    column = i;
            }

            if (tableau[numberOfConstraints, column] <= 0)
                return -1; // optimalni
            else
                return column;
        }

        /// <summary>
		/// Metoda hledajici index klicoveho sloupce s nejvice negativni cenou - pivot column
		/// </summary>
		/// <returns>index klicoveho sloupce s nejvice negativni cenou nebo -1</returns>
        private int mini()
        {
            int column = 0;
            for (int i = 1; i < numberOfConstraints + numberOfOriginalVariables; i++)
            {
                if (tableau[numberOfConstraints, i] < tableau[numberOfConstraints, column])
                    column = i;
            }

            if (tableau[numberOfConstraints, column] >= 0)
                return -1; // optimalni
            else
                return column;
        }

        /// <summary>
		/// Metoda hledajici klicovy radek ("pivot row") pouzitim Pravidla Minimalniho Pomeru "min_ratio" (-1 pokud neni takovy radek)
		/// </summary>
		/// <returns>index klicoveho radku nebo -1</returns>
		/// <param name="columnIndex">index klicoveho sloupce - "pivot column"</param>
        private int minRatioRule(int columnIndex)
        {
            int row = -1;
            for (int i = 0; i < numberOfConstraints; i++)
            {
                if (tableau[i, columnIndex] <= 0)
                    continue;
                else if (row < 0)
                    row = i;
                else if ((tableau[i, numberOfConstraints + numberOfOriginalVariables] / tableau[i, columnIndex]) < 
                    (tableau[row, numberOfConstraints + numberOfOriginalVariables] / tableau[row, columnIndex]))
                    row = i;
            }
            return row;
        }

        /// <summary>
		/// Pivot - klicovy (radek, sloupec) pouzitim Gauss-Jordanovi eliminacni metody
		/// </summary>
		/// <param name="row">Radek</param>
		/// <param name="column">Sloupec</param>
        private void pivot(int row, int column)
        {

            // vse z radku p a sloupce q
            for (int i = 0; i <= numberOfConstraints; i++)
                for (int j = 0; j <= numberOfConstraints + numberOfOriginalVariables; j++)
                    if (i != row && j != column)
                        tableau[i, j] -= tableau[row, j] * tableau[i, column]
                          / tableau[row, column];

            // nuly mimo sloupec q
            for (int i = 0; i <= numberOfConstraints; i++)
                if (i != row)
                    tableau[i, column] = 0.0;

            // vahovy radek p
            for (int j = 0; j <= numberOfConstraints + numberOfOriginalVariables; j++)
                if (j != column)
                    tableau[row, j] /= tableau[row, column];
            tableau[row, column] = 1.0;
        }

        /// <summary>
		/// Optimalni cilova hodnota Z
		/// </summary>
        public double value()
        {
            return -tableau[numberOfConstraints, numberOfConstraints + numberOfOriginalVariables];
        }

        /// <summary>
		/// Vector reseni
		/// </summary>
        public double[] solutionVector()
        {
            double[] vector = new double[numberOfOriginalVariables];
            for (int i = 0; i < numberOfConstraints; i++)
                if (basis[i] < numberOfOriginalVariables)
                    vector[basis[i]] = tableau[i, numberOfConstraints + numberOfOriginalVariables];
            return vector;
        }

        /// <summary>
		/// Vypis krok
		/// </summary>
        public void printStep()
        {
            for (int i = 0; i <= numberOfConstraints; i++)
            {
                for (int j = 0; j <= numberOfConstraints + numberOfOriginalVariables; j++)
                {
                    Console.Write("{0:0.00}\t", tableau[i, j]);
                }
                Console.WriteLine();
            }
            Console.ForegroundColor = ConsoleColor.DarkGreen;
            Console.WriteLine("Z = " + value());
            Console.ForegroundColor = ConsoleColor.DarkYellow;
            for (int i = 0; i < numberOfConstraints; i++)
                if (basis[i] < numberOfOriginalVariables)
                    Console.WriteLine("x_"
                      + basis[i]
                      + " = "
                      + tableau[i, numberOfConstraints + numberOfOriginalVariables]);
            Console.WriteLine();
            Console.ResetColor();
        }
    }
}

